/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.ast;

import ru.ispras.castle.ast.basis.SyntaxNode;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Assignment statement.
 *
 * <p>Consists of defined {@link RangedVariable} target
 * variable and used value expression.</p>
 *
 * @see RangedVariable
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 */
public final class Assignment<TagT extends Enum<TagT>> extends SyntaxNode<TagT> implements UseDef {

  private static final String SEPARATOR = ":=";

  private RangedVariable left;

  private Node right;

  /**
   * Constructs an assignment node.
   *
   * @param tagType the enumeration type representing node tags.
   * @param tag the node tag.
   */
  public Assignment(final Class<TagT> tagType, final TagT tag) {
    super(tagType, tag, SyntaxNode.Kind.BASIC);
  }

  /**
   * Constructs object with the specified target variable and value expression.
   *
   * @param tagType the enumeration type representing node tags.
   * @param tag the node tag.
   * @param left defined variable
   * @param right used expression
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public Assignment(
      final Class<TagT> tagType, final TagT tag, final RangedVariable left, final Node right) {
    this(tagType, tag);

    InvariantChecks.checkNotNull(left);
    InvariantChecks.checkNotNull(right);

    this.left = left;
    this.right = right;
  }

  /**
   * Constructs object with the specified target non-ranged variable and value expression.
   *
   * @param tagType the enumeration type representing node tags.
   * @param tag the node tag.
   * @param var defined variable
   * @param expr used expression
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public Assignment(
      final Class<TagT> tagType, final TagT tag, final NodeVariable var, final Node expr) {
    this(tagType, tag);

    InvariantChecks.checkNotNull(var);
    InvariantChecks.checkNotNull(expr);

    this.left = new RangedVariable(var);
    this.right = expr;
  }

  private Assignment(final Assignment<TagT> assignment) {
    super(assignment);

    this.left = assignment.left.deepCopy();
    this.right = assignment.right.deepCopy();
  }

  /**
   * Returns defined ranged variable.
   *
   * @return ranged variable
   */
  public RangedVariable getLeft() {
    return this.left;
  }

  /**
   * Returns defined variable.
   *
   * @return variable
   */
  public NodeVariable getVariable() {
    return this.left.getVariable();
  }

  /**
   * Checks rather this object contains a defining to ranged variable.
   *
   * @return {@code true} if this object defines ranged variable,
   *         {@code false} otherwise
   */
  public boolean ranged() {
    return this.left.ranged();
  }

  /**
   * Sets the specified variable as objects' target.
   *
   * @param variable new target variable
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setVariable(final RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.left = variable;
  }

  /**
   * Returns value expression.
   *
   * @return used expression
   */
  public Node getRight() {
    return this.right;
  }

  /**
   * Sets new value expression.
   *
   * @param right new value expression
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setRight(final Node right) {
    InvariantChecks.checkNotNull(right);

    this.right = right;
  }

  /**
   * Returns names of variables that are used in this object.
   *
   * @return set of used variables' names.
   */
  public Set<String> getUsedVariableNames() {
    final Collection<NodeVariable> usedVariables = getUses();
    final Set<String> usedVariableNames = new LinkedHashSet<>();

    for (final NodeVariable variable: usedVariables) {
      usedVariableNames.add(variable.getName());
    }
    return usedVariableNames;
  }

  /**
   * Returns this assignment description that is presented in format of the specified printer
   * with the specified separator.
   *
   * @param printer expression printer
   * @param separator assignment separator between defined and used variables
   *
   * @return description
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public String getSeparatedDescription(final ExprTreePrinter printer, final String separator) {
    InvariantChecks.checkNotNull(printer);
    InvariantChecks.checkNotNull(separator);

    return this.left.toString() + separator + printer.toString(this.right);
  }

  /**
   * Checks rather all the object's fields are not {@code null}.
   *
   * @return {@code true} when all the object's fields are not {@code null},
   *         {@code false} otherwise
   */
  public boolean isInitialized() {
    return this.left != null && this.right != null;
  }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> uses;
    if (this.right == null) {
      uses = new LinkedHashSet<>();
    } else {
      uses = new LinkedHashSet<>(ExprUtils.getVariables(this.right));
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    final Set<NodeVariable> defines = new LinkedHashSet<>();

    if (getLeft() != null) {
      defines.add(getVariable());
    }

    return defines;
  }

  @Override
  public Assignment<TagT> clone() {
    return new Assignment<>(this);
  }

  @Override
  public int hashCode() {
    int prime = 31;
    int result = 1;
    result = prime * result + ((this.right == null) ? 0 : this.right.hashCode());
    result = prime * result + this.left.hashCode();

    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }

    if (obj == null || !(obj instanceof Assignment)) {
      return false;
    }

    final Assignment other = (Assignment) obj;

    if (this.right == null) {
      if (other.right != null) {
        return false;
      }
    } else if (!this.right.equals(other.right)) {
      return false;
    }

    return this.left.equals(other.left);
  }

  @Override
  public String toString() {
    return this.left + " " + SEPARATOR + " " + this.right;
  }
}
