/*
 * Copyright 2013-2020 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.ast;

import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Variable or it's sub-part restricted by the specified range and indices.
 *
 * <p>When range or indices are not {@code null} it means that not all variable is in use
 * but only it's sub-part.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class RangedVariable implements UseDef {

  private final NodeVariable variable;

  /* Bit vector sub-range. */
  private Range range;

  /* Array element indices. */
  private final List<Node> indices;

  /**
   * Constructs object with the specified variable.
   *
   * @param variable The target variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public RangedVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.variable = variable;
    this.range = null;
    this.indices = new LinkedList<>();
  }

  /**
   * Constructs object with the specified variable and it's range.
   *
   * <p>The ranged variable MUST have either bit vector data type
   * or map data type.</p>
   *
   * @param variable The target variable.
   * @param range The variable's sub-range.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null};
   */
  public RangedVariable(final NodeVariable variable, final Range range) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(range);

    this.variable = variable;
    this.range = range;
    this.indices = new LinkedList<>();
  }

  /**
   * Creates object with the specified variable and it's range.
   *
   * <p>The indexed variable MUST be of map data type.</p>
   *
   * @param variable The target variable.
   * @param indices The list of indices describing the variable's element.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null};
   */
  public RangedVariable(final NodeVariable variable, final List<Node> indices) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(indices);

    this.variable = variable;
    this.range = null;
    this.indices = new LinkedList<>(indices);
  }

  /**
   * Creates object with the specified parameters.
   * @param var The target variable.
   * @param range The variable sub-range (for bit-vectors and bit-vector arrays).
   * @param indices The variable element indices (for arrays).
   */
  public RangedVariable(final NodeVariable var, final Range range, final List<Node> indices) {

    this.variable = var;
    this.range = range;
    this.indices = new LinkedList<>(indices);
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return Deep copy of this object.
   */
  public RangedVariable deepCopy() {

    final RangedVariable deepCopy = new RangedVariable((NodeVariable) this.variable.deepCopy());

    if (ranged()) {
      deepCopy.setRange(this.range.deepCopy());
    }

    if (indexed()) {

      final List<Node> indexCopies = new LinkedList<>();
      for (final Node index : getIndices()) {
        indexCopies.add(index.deepCopy());
      }
      deepCopy.setIndices(indexCopies);
    }
    return deepCopy;
  }

  /**
   * Returns variable.
   *
   * @return Variable.
   */
  public NodeVariable getVariable() {
    return this.variable;
  }

  /**
   * Returns variable name.
   *
   * @return Variable name.
   */
  public String getVariableName() {
    return this.getVariable().getName();
  }

  /**
   * Returns string form of the ranged variable that can be used as name.
   * @return The string name of the object.
   */
  public String getName() {

    final StringBuilder builder = new StringBuilder(getVariableName());

    if (indexed()) {
      for (final Node index : getIndices()) {
        builder.append("_").append(index.toString());
      }
    }

    if (ranged()) {
      final Node high = getRange().getHigh();
      final Node low = getRange().getLow();
      builder.append("_").append(high.toString());
      builder.append("_").append(low.toString());
    }

    return builder.toString();
  }
  /**
   * Returns range.
   *
   * @return Range.
   */
  public Range getRange() {
    return this.range;
  }

  /**
   * Returns the integer range of this object.
   *
   * @return The integer range of this object.
   */
  public IntegerRange getIntegerRange() {

    return getRange().toIntegerRange();
  }

  /**
   * Returns the sequence of object's indices.
   *
   * @return The sequence of object's indices.
   */
  public List<Node> getIndices() {

    return this.indices;
  }

  /**
   * Sets new range.
   *
   * @param range Range to be set.
   *
   *  @throws IllegalArgumentException when argument is {@code null};
   */
  public void setRange(final Range range) {
    InvariantChecks.checkNotNull(range);

    this.range = range;
  }

  /**
   * Sets new range from the pair of the specified nodes.
   * @param low The low border node.
   * @param high The high border node.
   */
  public void setRange(final Node low, final Node high) {
    setRange(new Range(low, high));
  }

  /**
   * Sets new sequence on indices.
   *
   * @param indices The sequence of indices.
   *
   *  @throws IllegalArgumentException in the following situations:
   *         (1) when any argument is {@code null};
   *         (2) when this object's variable is not of
   *         {@link ru.ispras.fortress.data.types.datamap.DataMap} data types;
   */
  public void setIndices(final List<Node> indices) {
    InvariantChecks.checkNotNull(indices);
    InvariantChecks.checkTrue(getVariable().isType(DataTypeId.MAP));

    this.indices.clear();
    this.indices.addAll(indices);
  }

  /**
   * Add the specified node as index for this object.
   * @param index The index.
   *
   *   @throws IllegalArgumentException in the following situations:
   *         (1) when any argument is {@code null};
   *         (2) when this object's variable is not of
   *         {@link ru.ispras.fortress.data.types.datamap.DataMap} data types;
   */
  public void addIndex(final Node index) {
    InvariantChecks.checkNotNull(index);
    InvariantChecks.checkTrue(getVariable().isType(DataTypeId.MAP));

    this.indices.add(index);
  }

  /**
   * Checks whether this object has sub-range.
   *
   * @return {@code true} if this contains non-{@code null} range, {@code false} otherwise.
   */
  public boolean ranged() {
    return this.range != null;
  }

  /**
   * Checks whether this object has indices.
   *
   * @return {@code true} if this object has indices, {@code false} otherwise.
   */
  public boolean indexed() {
    return !this.indices.isEmpty();
  }

  /**
   * Returns the syntax node that is equivalent to this object.
   * @return The node formula for this object.
   */
  public Node toNode() {

    Node result = getVariable();

    if (ranged()) {

      result = Nodes.bvextract(range.getHigh(), range.getLow(), result);
    }

    if (indexed()) {

      for (final Node index : getIndices()) {

        result = Nodes.select(result, index);
      }
    }

    return result;
  }

  @Override
  public String toString() {

    final StringBuilder builder = new StringBuilder();
    builder.append(this.variable.toString());

    if (indexed()) {

      for (final Node index : getIndices()) {
        builder.append(String.format("[%s]", index));
      }
    }
    if (ranged()) {

      builder.append(String.format("[%s]", this.range.toString()));
    }
    return builder.toString();
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final RangedVariable objVar = (RangedVariable) obj;

    if (!getVariable().equals(objVar.getVariable())) {
      return false;
    }
    if (getRange() == null ? objVar.getRange() != null : !getRange().equals(objVar.getRange())) {
      return false;
    }
    return getIndices() == null
        ? objVar.getIndices() == null
        : getIndices().equals(objVar.getIndices());
  }

  @Override
  public int hashCode() {
    int result = getVariable().hashCode();
    result = 31 * result + (getRange() == null ? 0 : getRange().hashCode());
    result = 31 * result + (getIndices() == null ? 0 : getIndices().hashCode());
    return result;
  }

  @Override
  public Set<NodeVariable> getUses() {

    final Set<NodeVariable> uses = new LinkedHashSet<>();
    uses.add(getVariable());

    if (ranged()) {

      uses.addAll(ExprUtils.getVariables(getRange().getHigh()));
      uses.addAll(ExprUtils.getVariables(getRange().getLow()));
    }

    if (indexed()) {
      uses.addAll(ExprUtils.getVariables(getIndices()));
    }

    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return Collections.emptySet();
  }
}
