/*
 * Copyright 2013-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.ast.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.Iterator;
import java.util.Stack;

/**
 * {@link SyntaxTreeWalker} implements a walker that traverses an abstract syntax tree.
 *
 * @see SyntaxNode
 * @see SyntaxNodeVisitor
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class SyntaxTreeWalker<TagT extends Enum<TagT>> {
  /**
   * {@link Entry} is a traversal stack entry.
   */
  private final class Entry {
    /** Node being traversed. */
    private final SyntaxNode<TagT> node;
    /** Iterator over the node's children. */
    private final Iterator<? extends SyntaxNode<TagT>> iterator;

    /**
     * Constructs a traversal stack entry.
     *
     * @param node the node to be traversed.
     * @param iterator the iterator over the node's children
     */
    public Entry(final SyntaxNode<TagT> node, final Iterator<? extends SyntaxNode<TagT>> iterator) {
      InvariantChecks.checkNotNull(node);
      InvariantChecks.checkNotNull(iterator);

      this.node = node;
      this.iterator = iterator;
    }
  }

  /** Root of the tree to be traversed. */
  private final SyntaxNode<TagT> root;
  /** User-defined visitor of tree nodes. */
  private final SyntaxNodeVisitor<TagT> visitor;

  /** Traversal stack. */
  private final Stack<Entry> stack = new Stack<>();

  /**
   * Constructs a tree walker.
   *
   * @param root the root of the tree to be traversed.
   * @param visitor the user-defined visitor of tree nodes.
   */
  public SyntaxTreeWalker(final SyntaxNode<TagT> root, final SyntaxNodeVisitor<TagT> visitor) {
    InvariantChecks.checkNotNull(root);
    InvariantChecks.checkNotNull(visitor);

    this.root = root;
    this.visitor = visitor;
  }

  /**
   * Starts traversal (depth-first exploration).
   */
  public final void start() {
    onBegin(root);

    while (!stack.isEmpty()) {
      final Entry entry = stack.peek();

      if (entry.iterator != null && entry.iterator.hasNext()) {
        onBegin(entry.iterator.next());
      } else {
        onEnd(entry.node);
      }
    }
  }

  /**
   * Starts visiting the node (calls the {@link SyntaxTreeWalker#onBegin} method).
   *
   * @param node the node to be visited.
   */
  private void onBegin(final SyntaxNode<TagT> node) {
    InvariantChecks.checkNotNull(node);

    switch (visitor.onBegin(node)) {
      case OK:
        stack.push(new Entry(node, node.getChildren().iterator()));
        break;
      case PRUNE:
        stack.push(new Entry(node, /* Do not traverse children */ null));
        break;
      case REPEAT:
        assert false;
        break;
      case ABORT:
        stack.clear();
        break;
    }
  }

  /**
   * Stops visiting the node (calls the {@link SyntaxTreeWalker#onBegin} method).
   *
   * @param node the node to be visited.
   */
  private void onEnd(final SyntaxNode<TagT> node) {
    InvariantChecks.checkNotNull(node);

    switch (visitor.onEnd(node)) {
      case OK:
        stack.pop();
        break;
      case PRUNE:
        assert false;
        break;
      case REPEAT:
        stack.pop();
        onBegin(node);
        break;
      case ABORT:
        stack.clear();
        break;
    }
  }
}
