/*
 * Copyright 2015-2020 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.util;

import ru.ispras.fortress.util.InvariantChecks;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;

/**
 * The {@link FileUtils} class provides utility methods to work with files.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class FileUtils {

  private FileUtils() {
    /* Empty. */
  }

  /**
   * Creates normalized file path.
   * <p>In normalized path non-system separators are substituted by system separators.</p>
   * @param filePath The path to file.
   * @return The new string that stores the normalized version of the input path.
   */
  public static String getNormalizedPath(final String filePath) {
    InvariantChecks.checkNotNull(filePath);

    if ('/' != File.separatorChar) {
      return filePath.replace('/', File.separatorChar);
    }
    if ('\\' != File.separatorChar) {
      return filePath.replace('\\', File.separatorChar);
    }

    return filePath;
  }

  /**
   * Returns extension of the file.
   * @param fileName The file name.
   * @return The file extension.
   */
  public static String getFileExtension(final String fileName) {
    InvariantChecks.checkNotNull(fileName);

    final int lastIndexOf = fileName.lastIndexOf('.');
    return lastIndexOf != -1 ? fileName.substring(lastIndexOf) : "";
  }

  /**
   * Returns the short name of the file.
   * @param fileName The file name.
   * @return The short name of the file.
   */
  public static String getShortFileName(final String fileName) {
    InvariantChecks.checkNotNull(fileName);
    return new File(fileName).getName();
  }

  /**
   * Returns the short name of the file without an extension.
   * @param fileName The file name.
   * @return The short name of the file without an extension.
   */
  public static String getShortFileNameNoExt(final String fileName) {
    InvariantChecks.checkNotNull(fileName);

    final String shortFileName = getShortFileName(fileName);
    final int dotPos = shortFileName.lastIndexOf('.');

    if (-1 == dotPos) {
      return shortFileName.toLowerCase();
    }

    return shortFileName.substring(0, dotPos).toLowerCase();
  }

  /**
   * Returns the parent directory of the file.
   * @param fileName The file name.
   * @return The parent directory of the file.
   */
  public static String getFileDir(final String fileName) {
    InvariantChecks.checkNotNull(fileName);

    final File file = new File(fileName);
    return file.getParent();
  }

  /**
   * Copies all data from the source location (file. directory, etc.) to the target one.
   * @param sourceLocation The source location.
   * @param targetLocation The target location.
   * @throws IOException when some internal file-related operation is crashed.
   */
  public static void copy(final File sourceLocation, final File targetLocation) throws IOException {
    if (sourceLocation.isDirectory()) {
      copyDirectory(sourceLocation, targetLocation);
    } else {
      copyFile(sourceLocation, targetLocation);
    }
  }

  /**
   * Copies all data from the source directory to the target directory.
   * @param source The source directory.
   * @param target The target directory.
   * @throws IOException when some internal file-related operation is crashed.
   */
  public static void copyDirectory(File source, File target) throws IOException {
    if (!target.exists()) {
      target.mkdir();
    }

    for (final String f : source.list()) {
      copy(new File(source, f), new File(target, f));
    }
  }

  /**
   * Copies all data from the source file to the target file.
   * @param source The source file.
   * @param target The target file.
   * @throws IOException when some internal file-related operation is crashed.
   */
  public static void copyFile(final File source, final File target) throws IOException {
    try (final InputStream in = new FileInputStream(source);
        final OutputStream out = new FileOutputStream(target)) {
      byte[] buf = new byte[1024];
      int length;
      while ((length = in.read(buf)) > 0) {
        out.write(buf, 0, length);
      }
    }
  }

  /**
   * Creates an empty file with the specified name at the specified path.
   * @param path The path to file.
   * @param name The file name without an extension.
   * @param ext The file extension.
   * @return The new file.
   */
  public static File newFile(
      final String path,
      final String name,
      final String ext) {
    InvariantChecks.checkNotNull(path);
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(ext);

    final File file = new File(path, name + "." + ext);
    createParentDirs(file);

    return file;
  }

  /**
   * Creates the new file with the specified name.
   * @param fileName The file name.
   * @return The new file with the specified name.
   */
  public static File newFile(final String fileName) {
    InvariantChecks.checkNotNull(fileName);

    final File file = new File(fileName);
    createParentDirs(file);

    return file;
  }

  /**
   * Checks whether the specified file is executable.
   * @param file The File to be checked.
   * @return {@code true} if the specified file exists and can be executed, {@code false} otherwise.
   */
  public static boolean isExecutable(final File file) {
    InvariantChecks.checkNotNull(file);

    return file.exists() && file.canExecute();
  }

  private static void createParentDirs(final File file) {
    InvariantChecks.checkNotNull(file);

    if (!file.exists()) {
      final File parent = file.getParentFile();
      if (null != parent && !parent.exists()) {
        parent.mkdirs();
      }
    }
  }

  /**
   * Creates new {@link PrintWriter} for the new file with the specified parameters.
   * @param path The path to new file.
   * @param name The new file name.
   * @param ext The new file extension.
   * @return The {@link PrintWriter} for the new file.
   * @throws IOException when some internal file-related operation is crashed.
   */
  public static PrintWriter newPrintWriter(
      final String path,
      final String name,
      final String ext) throws IOException {
    final File file = newFile(path, name, ext);
    return new PrintWriter(file);
  }

  /**
   * Creates new {@link BufferedReader} for existing file.
   *
   * @param file The processed file.
   * @return The {@link BufferedReader} for the file.
   * @throws FileNotFoundException when file does not exist.
   */
  public static BufferedReader newBufferedReader(final File file) {

    BufferedReader reader = null;
    try {
      reader = new BufferedReader(new FileReader(file));
    } catch (final FileNotFoundException e) {
      e.printStackTrace();
    }
    return reader;
  }
}
