/*
 * Copyright 2015-2019 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.castle.util;

/**
 * The {@link Logger} class provides methods for logging-related tasks.
 *
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public final class Logger {

  /**
   * Log message types.
   */
  public enum EventType {
    DEBUG    (true,  ""),
    MESSAGE  (false, ""),
    WARNING  (false, "Warning: "),
    ERROR    (false, "Error: ");

    private final boolean isDebugOnly;
    private final String textPrefix;

    EventType(final boolean isDebugOnly, final String textPrefix) {
      this.isDebugOnly = isDebugOnly;
      this.textPrefix = textPrefix;
    }
  }

  /**
   * The listener for logger.
   */
  public interface Listener {
    void onEventLogged(Logger.EventType type, String message);
  }


  /**
   * The log line width.
   */
  public static final int LINE_WIDTH = 80;

  /**
   * The separation line.
   */
  public static final String BAR = makeBar('-', LINE_WIDTH);

  private static boolean isDebug = false;
  private static Listener listener = null;
  private static String supportEmail = null;

  /**
   * Enables\disables debug mode.
   * @param value {@code true} for debug mode to be enabled, {@code false} otherwise.
   */
  public static void setDebug(final boolean value) {
    isDebug = value;
  }

  /**
   * Checks whether debug mode is enabled, or not.
   * @return {@code true} if debug mode is enabled, {@code false} otherwise.
   */
  public static boolean isDebug() {
    return isDebug;
  }

  /**
   * Sets the associated listener.
   * @param value The listener that should be associated to the logger.
   */
  public static void setListener(final Listener value) {
    listener = value;
  }

  /**
   * Sets the support email.
   * @param email The support email string.
   */
  public static void setSupportEmail(final String email) {
    supportEmail = email;
  }

  /**
   * Returns the support email string.
   * @return The support email string.
   */
  public static String getSupportEmail() {
    return supportEmail;
  }

  /**
   * Prints formatted message with the specified arguments to the log.
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void message(final String format, final Object... args) {
    print(EventType.MESSAGE, format, args);
  }

  /**
   * Prints formatted warning with the specified arguments to the log.
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void warning(final String format, final Object... args) {
    print(EventType.WARNING, format, args);
  }

  /**
   * Prints formatted message with the specified arguments to the log.
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void error(final String format, final Object... args) {
    print(EventType.ERROR, format, args);
  }

  /**
   * Prints formatted debug message with the specified arguments to the log.
   * @param format The message format.
   * @param args The arguments to be printed.
   */
  public static void debug(final String format, final Object... args) {
    print(EventType.DEBUG, format, args);
  }

  /**
   * If debug mode is on, prints separation line.
   * <p>If debug mode is off, does nothing.</p>
   */
  public static void debugBar() {
    if (isDebug) {
      print(EventType.DEBUG, BAR);
    }
  }

  /**
   * When debug mode is on, prints formatted arguments surrounded by debug header.
   * <p>When debug mode is off, does nothing.</p>
   * @param format The format for arguments.
   * @param args The arguments to be printed.
   */
  public static void debugHeader(final String format, final Object... args) {
    if (isDebug) {
      debugHeader(null != format ? String.format(format, args) : null);
    }
  }

  /**
   * When debug mode is on, surrounds the text with debug header.
   * <p>When debug mode is off, does nothing.</p>
   * @param text The text to be printed.
   */
  public static void debugHeader(final String text) {
    if (!isDebug) {
      return;
    }

    if (null == text || text.isEmpty()) {
      print(EventType.DEBUG, System.lineSeparator() + BAR + System.lineSeparator());
      return;
    }

    final int barLength = LINE_WIDTH - text.length();
    final int prefixLength  = barLength > 0 ? barLength / 2 : 0;
    final int postfixLength = barLength > 0 ? barLength - prefixLength : 0;

    final StringBuilder sb = new StringBuilder();

    sb.append(System.lineSeparator());
    sb.append(makeBar('-', prefixLength - 1));
    sb.append(' ');
    sb.append(text);
    sb.append(' ');
    sb.append(makeBar('-',  postfixLength - 1));
    sb.append(System.lineSeparator());

    print(EventType.DEBUG, sb.toString());
  }

  /**
   * Prints error message for the unexpected event.
   * @param throwable The unexpected event.
   */
  public static void exception(final Throwable throwable) {
    final StringBuilder sb = new StringBuilder();

    sb.append(makeBar('*', LINE_WIDTH));
    sb.append(System.lineSeparator());
    sb.append("ATTENTION! An unexpected error has occurred:");
    sb.append(System.lineSeparator());

    if (throwable != null) {
      sb.append(String.format("%s: %s", throwable.getClass().getName(), throwable.getMessage()));
    }
    sb.append(System.lineSeparator());

    sb.append(System.lineSeparator());
    sb.append("The program will be terminated. Please contact us at: ");
    sb.append(System.lineSeparator());
    sb.append(getSupportEmail());

    sb.append(System.lineSeparator());
    sb.append("We are sorry for the inconvenience.");

    sb.append(System.lineSeparator());
    sb.append(System.lineSeparator());
    sb.append("Exception stack:");
    sb.append(System.lineSeparator());
    sb.append(System.lineSeparator());

    final java.io.StringWriter writer = new java.io.StringWriter();
    throwable.printStackTrace(new java.io.PrintWriter(writer));
    sb.append(writer.toString());

    sb.append(System.lineSeparator());
    sb.append(makeBar('*', LINE_WIDTH));

    print(Logger.EventType.ERROR, sb.toString());
  }

  private static void print(
      final EventType type, final String format, final Object... args) {

    if (type.isDebugOnly && !isDebug) {
      return;
    }

    if (null == args || 0 == args.length) {
      print(type, type.textPrefix + format);
      return;
    }

    print(type, type.textPrefix + (null != format ? String.format(format, args) : null));
  }

  private static void print(final EventType type, final String text) {
    System.out.println(text);

    if (null != listener) {
      listener.onEventLogged(type, text);
    }
  }

  private static String makeBar(final char ch, final int length) {
    if (length <= 0) {
      return "";
    }

    final StringBuilder sb = new StringBuilder(length);
    for (int i = 0; i < length; ++i) {
      sb.append(ch);
    }
    return sb.toString();
  }
}
